#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/times.h>
#if defined(sun)
#include <sys/param.h>
#endif
#include "align.h"

alignment_t *astar_pairopt(int dim, alignment_t *alignment,
			   int wk, int fs_k, int bs_k, int dir_mask)
{
  estimator_t *fs_estimator, *bs_estimator;
  pqueue_t fs_pqueue, bs_pqueue;
  node_set_t node_set;
  node_t node, next_node, mid_node, orig_node, dest_node;
  node_list_t bridge_node_list, node_list;
  int max_score_multiplied;
  int gd;
  int sub, edge_cost, add_cost, al_len, al_score, est_cost;
  int i, j, k, l, m, n, p, q, is_bridge;
  int next_pos[MAX_DIM], pos[MAX_DIM];
  alignment_t *al_p;
#ifdef VERBOSE
  double wnodes;
  int fs_visited, bs_visited, fs_expanded, bs_expanded;
  struct tms t_start, t_search_start, t_search_end, t_end;
#endif

#ifdef VERBOSE
  times(&t_start);
#endif
  if(dim <= 0){
    return NULL;
  }else if(dim == 1){
    return alignment;
  }

  mul_score();
  gd = (gap_denominator != 1);

  n = 0;
  for(i = 0; i < dim; i++){
/*    if(alignment[i].seq[0].len > n){ */
    if(alignment[i].dim > n){
      n = alignment[i].dim;
    }
  }

  max_score_multiplied = max_score * gap_denominator;
  if(internal_gap_penalty > max_score_multiplied)
    max_score_multiplied = internal_gap_penalty;
  if(external_gap_penalty > max_score_multiplied)
    max_score_multiplied = external_gap_penalty;

  add_cost = n * n * max_score_multiplied;

  if(dir_mask & FMASK){
    if((fs_estimator
	= (estimator_t *)malloc(sizeof(estimator_t) * dim * (dim - 1) / 2))
       == NULL){
      fprintf(stderr, "Can't allocate memory for fs-estimator\n");
      exit(-1);
    }
    sub = 0;
    for(i = 0; i < dim; i++){
      for(j = i + 1; j < dim; j ++){
#ifdef VERBOSE
	fprintf(stderr,	"Calculating estimates (forward search) %d:%d .... ",
		i, j);
#endif
	fs_estimator[sub++]
	  = estimator_init(&(alignment[i]), &(alignment[j]), FORWARD);
#ifdef VERBOSE
	fprintf(stderr, "done\n");
#endif
      }
    }
  }
  if(dir_mask & BMASK){
    if((bs_estimator
	= (estimator_t *)malloc(sizeof(estimator_t) * dim * (dim - 1) / 2))
       == NULL){
      fprintf(stderr, "Can't allocate memory for bs-estimator\n");
      exit(-1);
    }
    sub = 0;
    for(i = 0; i < dim; i++){
      for(j = i + 1; j < dim; j ++){
#ifdef VERBOSE
	fprintf(stderr, "Calculating estimates (backward search) %d:%d .... ",
		i, j);
#endif
	bs_estimator[sub++]
	  = estimator_init(&(alignment[i]), &(alignment[j]), BACKWARD);
#ifdef VERBOSE
	fprintf(stderr, "done\n");
#endif
      }
    }
  }
#ifdef VERBOSE
  times(&t_search_start);
  fs_visited = 0;
  bs_visited = 0;
  fs_expanded = 0;
  bs_expanded = 0;
#endif
  node_set = node_set_init(dim);
  for(i = 0; i < dim; i++){
    pos[i] = 0;
  }
  orig_node = create_node(node_set, pos);
  fs_info_init(orig_node);
  for(i = 0; i < dim; i++){
    pos[i] = alignment[i].seq[0].len;
  }
  dest_node = create_node(node_set, pos);
  bs_info_init(dest_node);
  bridge_node_list = node_list_init();
  if(dir_mask & FMASK){
    fs_pqueue = pqueue_init(FORWARD);
    est_cost = 0;
    sub = 0;
    for(i = 0; i < dim; i++){
      for(j = i + 1; j < dim; j ++){
	est_cost += fs_k * (lookup_estimate(fs_estimator[sub], 0, 0)
			    + add_cost * (alignment[i].seq[0].len
					  + alignment[j].seq[0].len));
	sub++;
      }
    }
    set_fs_cost(orig_node, est_cost);
    set_fs_status(orig_node, VISITED);
    put_node_to_pqueue(fs_pqueue, orig_node);
#ifdef VERBOSE
    fs_visited++;
#endif
  }else{
    set_fs_cost(orig_node, 0);
  }
  if(dir_mask & BMASK){
    bs_pqueue = pqueue_init(BACKWARD);
    est_cost = 0;
    sub = 0;
    for(i = 0; i < dim; i++){
      for(j = i + 1; j < dim; j ++){
	est_cost += bs_k * (lookup_estimate(bs_estimator[sub],
					    alignment[i].seq[0].len,
					    alignment[j].seq[0].len)
			    + add_cost * (alignment[i].seq[0].len
					  + alignment[j].seq[0].len));
	sub++;
      }
    }
    set_bs_cost(dest_node, est_cost);
    set_bs_status(dest_node, VISITED);
    put_node_to_pqueue(bs_pqueue, dest_node);
#ifdef VERBOSE
    bs_visited++;
#endif
  }else{
    set_bs_cost(dest_node, 0);
  }
  while(1){
    if(dir_mask & FMASK){
      node = get_node_from_pqueue(fs_pqueue);
      set_fs_status(node, FINISHED);
#ifdef VERBOSE
      fs_expanded++;
#endif
      if((node == dest_node) || is_bs_finished(node)){
	mid_node = node;
	break;
      }
      is_bridge = 0;
      for(i = 1; i < (1 << dim); i++){
	n = 1;
	for(j = 0; j < dim; j++){
	  if(i & n){
	    if((next_pos[j] = node->pos[j] + 1) > alignment[j].seq[0].len){
	      goto fs_next_edge;
	    }
	  }else{
	    next_pos[j] = node->pos[j];
	  }
	  n <<= 1;
	}
	edge_cost = 0;
	sub = 0;
	for(j = 0; j < dim; j++){
	  for(k = j + 1; k < dim; k ++){
	    if((node->pos[j] != next_pos[j]) && (node->pos[k] != next_pos[k])){
	      for(l = 0; l < alignment[j].dim; l++){
		for(m = 0; m < alignment[k].dim; m++){
		  edge_cost += wk
		    * matching_cost(alignment[j].seq[l].str[node->pos[j]],
				    alignment[k].seq[m].str[node->pos[k]]);
		}
	      }
	      edge_cost += wk * add_cost * 2;
	    }else if(node->pos[j] != next_pos[j]){
	      n = 0; 
	      for(l = 0; l < alignment[j].dim; l++){
		if(is_normal_char(alignment[j].seq[l].str[node->pos[j]])){
		  n++;
		}
	      }
	      for(m = 0; m < alignment[k].dim; m++){
		if(((node->pos[k] == 0)
		    || (alignment[k].seq[m].str[node->pos[k] - 1]
			== EXTERNAL_GAP_CHAR))
		   || ((node->pos[k] == alignment[k].seq[0].len)
		       || (alignment[k].seq[m].str[node->pos[k]]
			   == EXTERNAL_GAP_CHAR))){
		  edge_cost += wk * n * (-external_gap_penalty);
		}else{
		  edge_cost += wk * n * (-internal_gap_penalty);
		}
	      }
	      edge_cost += wk * add_cost;
	    }else if(node->pos[k] != next_pos[k]){
	      n = 0;
	      for(m = 0; m < alignment[k].dim; m++){
		if(is_normal_char(alignment[k].seq[m].str[node->pos[k]])){
		  n++;
		}
	      }
	      for(l = 0; l < alignment[j].dim; l++){
		if(((node->pos[j] == 0)
		    || (alignment[j].seq[l].str[node->pos[j] - 1]
			== EXTERNAL_GAP_CHAR))
		   || ((node->pos[j] == alignment[j].seq[0].len)
		       || (alignment[j].seq[l].str[node->pos[j]]
			   == EXTERNAL_GAP_CHAR))){
		  edge_cost += wk * n * (-external_gap_penalty);
		}else{
		  edge_cost += wk * n * (-internal_gap_penalty);
		}
	      }
	      edge_cost += wk * add_cost;
	    }
	    edge_cost
	      += fs_k *	(lookup_estimate(fs_estimator[sub],
					 next_pos[j], next_pos[k])
			 - lookup_estimate(fs_estimator[sub],
					   node->pos[j], node->pos[k])
			 - add_cost * ((next_pos[j] - node->pos[j])
				       + (next_pos[k] - node->pos[k])));
	    if(dir_mask & BMASK){
	      edge_cost
		+= bs_k * (lookup_estimate(bs_estimator[sub],
					   node->pos[j], node->pos[k])
			   - lookup_estimate(bs_estimator[sub],
					     next_pos[j], next_pos[k])
			   - add_cost * ((next_pos[j] - node->pos[j])
					 + (next_pos[k] - node->pos[k])));
	    }
	    sub++;
	  }
	}
	if((next_node = get_node(node_set, next_pos)) == NULL){
	  next_node = create_node(node_set, next_pos);
	}
	if(fs_info(next_node) == NULL){
	  fs_info_init(next_node);
	}
	if(is_bs_finished(next_node)){
	  is_bridge = 1;
	}
#ifdef VERBOSE
	if(lookup_fs_status(next_node) == UNVISITED){
	  fs_visited++;
	  if((fs_visited & 0x7fff) == 0){
	    fprintf(stderr,
		    ".... visited %d nodes (forward search)\n", fs_visited);
	  }
	}
#endif
	extend_node(fs_pqueue, node, next_node, edge_cost);
      fs_next_edge:
      }
      if(is_bridge){
	bridge_node_list = node_list_cons(node, bridge_node_list);
      }
    }
    if(dir_mask & BMASK){
      node = get_node_from_pqueue(bs_pqueue);
      set_bs_status(node, FINISHED);
#ifdef VERBOSE
      bs_expanded++;
#endif
      if((node == orig_node) || is_fs_finished(node)){
	mid_node = node;
	break;
      }
      for(i = 1; i < (1 << dim); i++){
	n = 1;
	for(j = 0; j < dim; j++){
	  if(i & n){
	    if((next_pos[j] = node->pos[j] - 1) < 0){
	      goto bs_next_edge;
	    }
	  }else{
	    next_pos[j] = node->pos[j];
	  }
	  n <<= 1;
	}
	edge_cost = 0;
	sub = 0;
	for(j = 0; j < dim; j++){
	  for(k = j + 1; k < dim; k ++){
	    if((node->pos[j] != next_pos[j]) && (node->pos[k] != next_pos[k])){
	      for(l = 0; l < alignment[j].dim; l++){
		for(m = 0; m < alignment[k].dim; m++){
		  edge_cost += wk
		    * matching_cost(alignment[j].seq[l].str[next_pos[j]],
				    alignment[k].seq[m].str[next_pos[k]]);
		}
	      }
	      edge_cost += wk * add_cost * 2;
	    }else if(node->pos[j] != next_pos[j]){
	      n = 0; 
	      for(l = 0; l < alignment[j].dim; l++){
		if(is_normal_char(alignment[j].seq[l].str[next_pos[j]])){
		  n++;
		}
	      }
	      for(m = 0; m < alignment[k].dim; m++){
		if(((next_pos[k] == 0)
		    || (alignment[k].seq[m].str[next_pos[k] - 1]
			== EXTERNAL_GAP_CHAR))
		   || ((next_pos[k] == alignment[k].seq[0].len)
		       || (alignment[k].seq[m].str[next_pos[k]]
			   == EXTERNAL_GAP_CHAR))){
		  edge_cost += wk * n * (-external_gap_penalty);
		}else{
		  edge_cost += wk * n * (-internal_gap_penalty);
		}
	      }
	      edge_cost += wk * add_cost;
	    }else if(node->pos[k] != next_pos[k]){
	      n = 0;
	      for(m = 0; m < alignment[k].dim; m++){
		if(is_normal_char(alignment[k].seq[m].str[next_pos[k]])){
		  n++;
		}
	      }
	      for(l = 0; l < alignment[j].dim; l++){
		if(((next_pos[j] == 0)
		    || (alignment[j].seq[l].str[next_pos[j] - 1]
			== EXTERNAL_GAP_CHAR))
		   || ((next_pos[j] == alignment[j].seq[0].len)
		       || (alignment[j].seq[l].str[next_pos[j]]
			   == EXTERNAL_GAP_CHAR))){
		  edge_cost += wk * n * (-external_gap_penalty);
		}else{
		  edge_cost += wk * n * (-internal_gap_penalty);
		}
	      }
	      edge_cost += wk * add_cost;
	    }
	    if(dir_mask & FMASK){
	      edge_cost
		+= fs_k * (lookup_estimate(fs_estimator[sub],
					   node->pos[j], node->pos[k])
			   - lookup_estimate(fs_estimator[sub],
					     next_pos[j], next_pos[k])
			   - add_cost * ((node->pos[j] - next_pos[j])
					 + (node->pos[k] - next_pos[k])));
	    }
	    edge_cost
	      += bs_k * (lookup_estimate(bs_estimator[sub],
					     next_pos[j], next_pos[k])
			 - lookup_estimate(bs_estimator[sub],
					   node->pos[j], node->pos[k])
			 - add_cost * ((node->pos[j] - next_pos[j])
				       + (node->pos[k] - next_pos[k])));
	    sub++;
	  }
	}
	if((next_node = get_node(node_set, next_pos)) == NULL){
	  next_node = create_node(node_set, next_pos);
	}
	if(bs_info(next_node) == NULL){
	  bs_info_init(next_node);
	}
	if(is_fs_finished(next_node)){
	  bridge_node_list = node_list_cons(next_node, bridge_node_list);
	}
#ifdef VERBOSE
	if(lookup_bs_status(next_node) == UNVISITED){
	  bs_visited++;
	  if((bs_visited & 0x7fff) == 0){
	    fprintf(stderr,
		    ".... visited %d nodes (backward search)\n", bs_visited);
	  }
	}
#endif
	extend_node(bs_pqueue, node, next_node, edge_cost);
      bs_next_edge:
      }
    }
  }
#ifdef VERBOSE
  times(&t_search_end);
#endif
  if(dir_mask & FMASK){
    sub = 0;
    for(i = 0; i < dim; i++){
      for(j = i + 1; j < dim; j++){
	estimator_finish(fs_estimator[sub]);
	sub++;
      }
    }
    pqueue_finish(fs_pqueue);
  }
  if(dir_mask & BMASK){
    sub = 0;
    for(i = 0; i < dim; i++){
      for(j = i + 1; j < dim; j++){
	estimator_finish(bs_estimator[sub]);
	sub++;
      }
    }
    pqueue_finish(bs_pqueue);
  }
  for(node_list = bridge_node_list;
      node_list != NULL; node_list = node_list->next){
    if(lookup_fs_cost(node_list->node) + lookup_bs_cost(node_list->node)
       < lookup_fs_cost(mid_node) + lookup_bs_cost(mid_node)){
      mid_node = node_list->node;
    }
  }
  if((al_p = (alignment_t *)malloc(sizeof(alignment_t))) == NULL){
    fprintf(stderr, "Can't allocate memory for result alignment");
    exit(-1);
  }
  al_p->dim = 0;
  for(i = 0; i < dim; i++){
    al_p->dim += alignment[i].dim;
  }
  al_len = 0;
  for(node = mid_node; node != orig_node; node = lookup_fs_prev_node(node)){
    al_len++;
  }
  for(node = mid_node; node != dest_node; node = lookup_bs_prev_node(node)){
    al_len++;
  }
  if((al_p->seq = (sequence_t *)malloc(sizeof(sequence_t) * al_p->dim))
     == NULL){
    fprintf(stderr, "Can't allocate memory for result alignment");
    exit(-1);
  }
  k = 0;
  for(i = 0; i < dim; i++){
    for(j = 0; j < alignment[i].dim; j++){
      al_p->seq[k].len = al_len;
      if((al_p->seq[k].str
	  = (char *)malloc(sizeof(char) * (al_len + 1))) == NULL){
	fprintf(stderr, "Can't allocate memory for result alignment");
	exit(-1);
      }
      al_p->seq[k].str[al_len] = '\0';
      p = -1;
      for(node = mid_node; node != orig_node;
	  node = lookup_fs_prev_node(node)){
	p++;
      }
      q = mid_node->pos[i] - 1;
      for(node = mid_node; node != orig_node;
	  node = lookup_fs_prev_node(node)){
	if(node->pos[i] != lookup_fs_prev_node(node)->pos[i]){
	  al_p->seq[k].str[p--] = alignment[i].seq[j].str[q--];
	}else{
	  if(((q + 1 > alignment[i].seq[0].len - 1)
	      || (alignment[i].seq[j].str[q + 1] == EXTERNAL_GAP_CHAR))
	     || ((q < 0)
		 || (alignment[i].seq[j].str[q] == EXTERNAL_GAP_CHAR))){
	    al_p->seq[k].str[p--] = EXTERNAL_GAP_CHAR;
	  }else{
	    al_p->seq[k].str[p--] = INTERNAL_GAP_CHAR;
	  }
	}
      }
      p = al_len;
      for(node = mid_node; node != dest_node;
	  node = lookup_bs_prev_node(node)){
	p--;
      }
      q = mid_node->pos[i];
      for(node = mid_node; node != dest_node;
	  node = lookup_bs_prev_node(node)){
	if(node->pos[i] != lookup_bs_prev_node(node)->pos[i]){
	  al_p->seq[k].str[p++] = alignment[i].seq[j].str[q++];
	}else{
	  if(((q > alignment[i].seq[0].len - 1)
	      || (alignment[i].seq[j].str[q] == EXTERNAL_GAP_CHAR))
	     || ((q - 1 < 0)
		 || (alignment[i].seq[j].str[q - 1] == EXTERNAL_GAP_CHAR))){
	    al_p->seq[k].str[p++] = EXTERNAL_GAP_CHAR;
	  }else{
	    al_p->seq[k].str[p++] = INTERNAL_GAP_CHAR;
	  }
	}
      }
      k++;
    }
  }
  al_score = -((lookup_fs_cost(mid_node) + lookup_bs_cost(mid_node)) / wk);
#ifdef VERBOSE
  fprintf(stderr, "Estimated inter-group cost: %d",
	  (lookup_fs_cost(orig_node) + lookup_bs_cost(dest_node)) / wk);
  if (gd) fprintf(stderr,"/%d", gap_denominator);
  fprintf(stderr,"\n");
  fprintf(stderr, "Actual inter-group cost: %d", -al_score);
  if (gd) fprintf(stderr,"/%d", gap_denominator);
  fprintf(stderr,"\n");
#endif
#ifndef NO_ITER
  node_set_finish(node_set);
#endif
  for(i = 0; i < dim; i++){
    for(j = i + 1; j < dim; j++){
      al_score
	+= add_cost * (alignment[i].seq[0].len + alignment[j].seq[0].len);
    }
  }
  for(i = 0; i < dim; i++){
    for(j = 0; j < alignment[i].dim; j++){
      for(k = j + 1; k < alignment[i].dim; k++){
	for(l = 0; l < alignment[i].seq[0].len; l++){
	  al_score += -matching_cost(alignment[i].seq[j].str[l],
				     alignment[i].seq[k].str[l]);
	}
      }
    }
  }
  al_p->score = al_score;
#ifdef VERBOSE
  times(&t_end);
  wnodes = alignment[0].seq[0].len + 1;
  for(i = 1; i < dim; i++){
    wnodes *= alignment[i].seq[0].len + 1;
  }
  fprintf(stderr, "# of whole nodes: %8.2e\n", wnodes);
  fprintf(stderr, "# of visited nodes (forward search): %d\n", fs_visited);
  fprintf(stderr, "# of expanded nodes (forward search): %d\n", fs_expanded);
  fprintf(stderr, "# of visited nodes (backward search): %d\n", bs_visited);
  fprintf(stderr, "# of expanded nodes (backward search): %d\n", bs_expanded);
  fprintf(stderr, "Used time:     User    System  Total\n");
#if defined(linux)
  fprintf(stderr, "  Pre-process  %6.2f  %6.2f  %6.2f\n",
	  (float)(t_search_start.tms_utime - t_start.tms_utime) / CLK_TCK,
	  (float)(t_search_start.tms_stime - t_start.tms_stime) / CLK_TCK,
	  (float)((t_search_start.tms_utime - t_start.tms_utime)
		  + (t_search_start.tms_stime - t_start.tms_stime)) / CLK_TCK);
  fprintf(stderr, "  Search       %6.2f  %6.2f  %6.2f\n",
	  (float)(t_search_end.tms_utime - t_search_start.tms_utime) / CLK_TCK,
	  (float)(t_search_end.tms_stime - t_search_start.tms_stime) / CLK_TCK,
	  (float)((t_search_end.tms_utime - t_search_start.tms_utime)
		  + (t_search_end.tms_stime - t_search_start.tms_stime))
	  / CLK_TCK);
  fprintf(stderr, "  Post-process %6.2f  %6.2f  %6.2f\n",
	  (float)(t_end.tms_utime - t_search_end.tms_utime) / CLK_TCK,
	  (float)(t_end.tms_stime - t_search_end.tms_stime) / CLK_TCK,
	  (float)((t_end.tms_utime - t_search_end.tms_utime)
		  + (t_end.tms_stime - t_search_end.tms_stime)) / CLK_TCK);
  fprintf(stderr, "  Total        %6.2f  %6.2f  %6.2f\n",
	  (float)(t_end.tms_utime - t_start.tms_utime) / CLK_TCK,
	  (float)(t_end.tms_stime - t_start.tms_stime) / CLK_TCK,
	  (float)((t_end.tms_utime - t_start.tms_utime)
		  + (t_end.tms_stime - t_start.tms_stime)) / CLK_TCK);
#elif defined(sun)
  fprintf(stderr, "  Pre-process  %6.2f  %6.2f  %6.2f\n",
	  (float)(t_search_start.tms_utime - t_start.tms_utime) / HZ,
	  (float)(t_search_start.tms_stime - t_start.tms_stime) / HZ,
	  (float)((t_search_start.tms_utime - t_start.tms_utime)
		  + (t_search_start.tms_stime - t_start.tms_stime)) / HZ);
  fprintf(stderr, "  Search       %6.2f  %6.2f  %6.2f\n",
	  (float)(t_search_end.tms_utime - t_search_start.tms_utime) / HZ,
	  (float)(t_search_end.tms_stime - t_search_start.tms_stime) / HZ,
	  (float)((t_search_end.tms_utime - t_search_start.tms_utime)
		  + (t_search_end.tms_stime - t_search_start.tms_stime)) / HZ);
  fprintf(stderr, "  Post-process %6.2f  %6.2f  %6.2f\n",
	  (float)(t_end.tms_utime - t_search_end.tms_utime) / HZ,
	  (float)(t_end.tms_stime - t_search_end.tms_stime) / HZ,
	  (float)((t_end.tms_utime - t_search_end.tms_utime)
		  + (t_end.tms_stime - t_search_end.tms_stime)) / HZ);
  fprintf(stderr, "  Total        %6.2f  %6.2f  %6.2f\n",
	  (float)(t_end.tms_utime - t_start.tms_utime) / HZ,
	  (float)(t_end.tms_stime - t_start.tms_stime) / HZ,
	  (float)((t_end.tms_utime - t_start.tms_utime)
		  + (t_end.tms_stime - t_start.tms_stime)) / HZ);
#else
  fprintf(stderr, "  Pre-process  %6d  %6d  %6d\n",
	  (int)(t_search_start.tms_utime - t_start.tms_utime),
	  (int)(t_search_start.tms_stime - t_start.tms_stime),
	  (int)((t_search_start.tms_utime - t_start.tms_utime)
		+ (t_search_start.tms_stime - t_start.tms_stime)));
  fprintf(stderr, "  Search       %6d  %6d  %6d\n",
	  (int)(t_search_end.tms_utime - t_search_start.tms_utime),
	  (int)(t_search_end.tms_stime - t_search_start.tms_stime),
	  (int)((t_search_end.tms_utime - t_search_start.tms_utime)
		+ (t_search_end.tms_stime - t_search_start.tms_stime)));
  fprintf(stderr, "  Post-process %6d  %6d  %6d\n",
	  (int)(t_end.tms_utime - t_search_end.tms_utime),
	  (int)(t_end.tms_stime - t_search_end.tms_stime),
	  (int)((t_end.tms_utime - t_search_end.tms_utime)
		+ (t_end.tms_stime - t_search_end.tms_stime)));
  fprintf(stderr, "  Total        %6d  %6d  %6d\n",
	  (int)(t_end.tms_utime - t_start.tms_utime),
	  (int)(t_end.tms_stime - t_start.tms_stime),
	  (int)((t_end.tms_utime - t_start.tms_utime)
		+ (t_end.tms_stime - t_start.tms_stime)));
#endif
#endif
  div_score();
  return al_p;
}
