#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/times.h>
#if defined(sun)
#include <sys/param.h>
#endif
#include "align.h"

void pairsubopt(int dim, alignment_t *alignment, int use_ub, int ubound,
		int wk, int fs_k, int bs_k, int dir_mask,
		int delta, int kbs, int line,
		int use_detour, int print_subopt,
		solutions_p opt_sols, solutions2_p opt_sols2,
		solutions2_p opt_sols3)
/* main routine */
{
  estimator_t *fs_estimator, *bs_estimator;
  pqueue_t fs_pqueue, bs_pqueue;
  node_set_t node_set;
  node_t node, next_node, mid_node, orig_node, dest_node;
  node_list_t bridge_node_list, node_list;
  int sub, edge_cost, add_cost, est_cost;
  int pair_edge_cost, pair_est;
  int max_score_multiplied;
  int al_len, al_score, intra_group_cost;
  int shortest_cost;
  int i, j, k, l, m, n, p, q, s, is_bridge;
  int next_pos[MAX_DIM], pos[MAX_DIM];
  alignment_t *al_p, *subopt_p;
  int tmp;
  int new_ubound, fs_farthest, bs_farthest;
  int *histogram;
  int depth, maxdepth;
  subopt_answer_t answer;
  int gd;
  int ig, eg, tg, is, es;
  int slope;
#ifdef VERBOSE
  double wnodes;
  int fs_visited, bs_visited, fs_expanded, bs_expanded;
  int fs_visited_for_opt, bs_visited_for_opt, fs_expanded_for_opt, bs_expanded_for_opt;
  int constructed[2], h_t[2], h_edge[2];
  struct tms t_start, t_search_start, t_optimal_end, t_search_end, t_eppstein_end, t_end;
  int total_sols;
#endif

#ifdef VERBOSE
  times(&t_start);
#endif

  if(dim <= 1) return;

  mul_score();
  gd = (gap_denominator != 1);

  n = 0;
  for(i = 0; i < dim; i++){
/*    if(alignment[i].seq[0].len > n){*/
    if(alignment[i].dim > n){
      n = alignment[i].dim;
    }
  }

#ifdef VERBOSE
  fprintf(stderr,"Using Gap Penalty: %d, %d / %d\n",internal_gap_penalty,external_gap_penalty,gap_denominator);
#endif

  max_score_multiplied = max_score * gap_denominator;
  if(internal_gap_penalty > max_score_multiplied)
    max_score_multiplied = internal_gap_penalty;
  if(external_gap_penalty > max_score_multiplied)
    max_score_multiplied = external_gap_penalty;

  add_cost = n * n * max_score_multiplied;
  if(dir_mask & FMASK){
    if((fs_estimator
	= (estimator_t *)malloc(sizeof(estimator_t) * dim * (dim - 1) / 2))
       == NULL){
      fprintf(stderr, "Can't allocate memory for fs-estimator\n");
      exit(-1);
    }

    sub = 0;
    for(i = 0; i < dim; i++){
      for(j = i + 1; j < dim; j ++){
#ifdef VERBOSE
	fprintf(stderr,	"Calculating estimates (forward search) %d:%d .... ",
		i, j);
#endif
	fs_estimator[sub++]
	  = estimator_init(&(alignment[i]), &(alignment[j]), FORWARD);
#ifdef VERBOSE
	fprintf(stderr, "done\n");
#endif
      }
    }
  }

  if(dir_mask & BMASK){
    if((bs_estimator
	= (estimator_t *)malloc(sizeof(estimator_t) * dim * (dim - 1) / 2))
       == NULL){
      fprintf(stderr, "Can't allocate memory for bs-estimator\n");
      exit(-1);
    }
    sub = 0;

    for(i = 0; i < dim; i++){
      for(j = i + 1; j < dim; j ++){
#ifdef VERBOSE
	fprintf(stderr, "Calculating estimates (backward search) %d:%d .... ",
		i, j);
#endif
	bs_estimator[sub++]
	  = estimator_init(&(alignment[i]), &(alignment[j]), BACKWARD);
#ifdef VERBOSE
	fprintf(stderr, "done\n");
#endif
      }
    }
  }

  intra_group_cost = 0;
  for(i = 0; i < dim; i++){
    for(j = 0; j < alignment[i].dim; j++){
      for(k = j + 1; k < alignment[i].dim; k++){
	for(l = 0; l < alignment[i].seq[0].len; l++){
	  intra_group_cost += matching_cost(alignment[i].seq[j].str[l],
					    alignment[i].seq[k].str[l]);
	}
      }
    }
  }

/*  delta *= gap_denominator;*/
  if (use_ub) {
/*    ubound *= gap_denominator;*/
    for(i = 0; i < dim; i++){
      for(j = i + 1; j < dim; j++){
	ubound
	  += add_cost * (alignment[i].seq[0].len + alignment[j].seq[0].len) * weights[i][j];
      }
    }
    ubound += delta - intra_group_cost;
  }

#ifdef VERBOSE
  if (use_ub) {
    fprintf(stderr, "Upper bound: %d", ubound);
    if (gd) fprintf(stderr, "/%d", gap_denominator);
    fprintf(stderr, "\n");
  }
  times(&t_search_start);
  fs_visited = 0;
  bs_visited = 0;
  fs_expanded = 0;
  bs_expanded = 0;
  constructed[0] = 0;
  constructed[1] = 0;
  h_t[0] = 0;
  h_t[1] = 0;
  h_edge[0] = 0;
  h_edge[1] = 0;
  total_sols = 0;
#endif
  node_set = node_set_init(dim);
  for(i = 0; i < dim; i++){
    pos[i] = 0;
  }
  orig_node = create_node(node_set, pos);
  fs_info_init(orig_node);
  for(i = 0; i < dim; i++){
    pos[i] = alignment[i].seq[0].len;
  }
  dest_node = create_node(node_set, pos);
  bs_info_init(dest_node);
  bridge_node_list = node_list_init();
  if(dir_mask & FMASK){
    fs_pqueue = pqueue_init(FORWARD);
    est_cost = 0;
    sub = 0;
    for(i = 0; i < dim; i++){
      for(j = i + 1; j < dim; j ++){
	est_cost += weights[i][j] * fs_k
	               * (lookup_estimate(fs_estimator[sub], 0, 0)
			  + add_cost * (alignment[i].seq[0].len
					+ alignment[j].seq[0].len));
	sub++;
      }
    }
    set_fs_cost(orig_node, est_cost);
    set_fs_status(orig_node, VISITED);
    put_node_to_pqueue(fs_pqueue, orig_node);
    fs_farthest = est_cost;
#ifdef VERBOSE
    fs_visited++;
#endif
  }else{
    set_fs_cost(orig_node, 0);
    fs_farthest = 0;
  }
  if(dir_mask & BMASK){
    bs_pqueue = pqueue_init(BACKWARD);
    est_cost = 0;
    sub = 0;
    for(i = 0; i < dim; i++){
      for(j = i + 1; j < dim; j ++){
	est_cost += weights[i][j] * bs_k *
	              (lookup_estimate(bs_estimator[sub],
				       alignment[i].seq[0].len,
				       alignment[j].seq[0].len)
		       + add_cost * (alignment[i].seq[0].len
				     + alignment[j].seq[0].len));
	sub++;
      }
    }
    set_bs_cost(dest_node, est_cost);
    set_bs_status(dest_node, VISITED);
    put_node_to_pqueue(bs_pqueue, dest_node);
    bs_farthest = est_cost;
#ifdef VERBOSE
    bs_visited++;
#endif
  }else{
    set_bs_cost(dest_node, 0);
    bs_farthest = 0;
  }
  while(1){
    if(dir_mask & FMASK){
      if((node = get_node_from_pqueue(fs_pqueue)) == NULL){
	fprintf(stderr, "Upper bound is too small\n");
	exit(-1);
      }
      set_fs_status(node, FINISHED);
      fs_farthest=lookup_fs_cost(node);
#ifdef VERBOSE
      fs_expanded++;
#endif
      is_bridge = 0;
      for(i = 1; i < (1 << dim); i++){
	n = 1;
	for(j = 0; j < dim; j++){
	  if(i & n){
	    if((next_pos[j] = node->pos[j] + 1) > alignment[j].seq[0].len){
	      goto fs_next_edge;
	    }
	  }else{
	    next_pos[j] = node->pos[j];
	  }
	  n <<= 1;
	}
	edge_cost = 0;
	est_cost = lookup_fs_cost(node);
	sub = 0;
	for(j = 0; j < dim; j++){
	  for(k = j + 1; k < dim; k ++){
	    pair_edge_cost = 0;
	    if((node->pos[j] != next_pos[j]) && (node->pos[k] != next_pos[k])){
	      for(l = 0; l < alignment[j].dim; l++){
		for(m = 0; m < alignment[k].dim; m++){
		  pair_edge_cost += weights[j][k] * wk
		    * matching_cost(alignment[j].seq[l].str[node->pos[j]],
				    alignment[k].seq[m].str[node->pos[k]]);
		}
	      }
	      pair_edge_cost += weights[j][k] * wk * add_cost * 2;
	    }else if(node->pos[j] != next_pos[j]){
	      n = 0; 
	      for(l = 0; l < alignment[j].dim; l++){
		if(is_normal_char(alignment[j].seq[l].str[node->pos[j]])){
		  n++;
		}
	      }
	      for(m = 0; m < alignment[k].dim; m++){
		if(((node->pos[k] == 0)
		    || (alignment[k].seq[m].str[node->pos[k] - 1]
			== EXTERNAL_GAP_CHAR))
		   || ((node->pos[k] == alignment[k].seq[0].len)
		       || (alignment[k].seq[m].str[node->pos[k]]
			   == EXTERNAL_GAP_CHAR))){
		  pair_edge_cost += weights[j][k] * wk * n * (-external_gap_penalty);
		}else{
		  pair_edge_cost += weights[j][k] * wk * n * (-internal_gap_penalty);
		}
	      }
	      pair_edge_cost += weights[j][k] * wk * add_cost;
	    }else if(node->pos[k] != next_pos[k]){
	      n = 0;
	      for(m = 0; m < alignment[k].dim; m++){
		if(is_normal_char(alignment[k].seq[m].str[node->pos[k]])){
		  n++;
		}
	      }
	      for(l = 0; l < alignment[j].dim; l++){
		if(((node->pos[j] == 0)
		    || (alignment[j].seq[l].str[node->pos[j] - 1]
			== EXTERNAL_GAP_CHAR))
		   || ((node->pos[j] == alignment[j].seq[0].len)
		       || (alignment[j].seq[l].str[node->pos[j]]
			   == EXTERNAL_GAP_CHAR))){
		  pair_edge_cost += weights[j][k] * wk * n * (-external_gap_penalty);
		}else{
		  pair_edge_cost += weights[j][k] * wk * n * (-internal_gap_penalty);
		}
	      }
	      pair_edge_cost += weights[j][k] * wk * add_cost;
	    }
	    pair_edge_cost
	      -= weights[j][k] * fs_k *
		         (lookup_estimate(fs_estimator[sub],
					  node->pos[j], node->pos[k])
			  + add_cost * ((alignment[j].seq[0].len
					 - node->pos[j])
					+ (alignment[k].seq[0].len
					   - node->pos[k])));
	    if(dir_mask & BMASK){
	      pair_edge_cost
		+= weights[j][k] * bs_k *
		      (lookup_estimate(bs_estimator[sub],
				       node->pos[j], node->pos[k])
		       + add_cost * (node->pos[j] + node->pos[k]));
	    }
	    pair_est
	      = lookup_estimate(fs_estimator[sub],
				next_pos[j], next_pos[k])
	      + add_cost * ((alignment[j].seq[0].len - next_pos[j])
			    + (alignment[k].seq[0].len - next_pos[k]));
	    est_cost += pair_edge_cost + weights[j][k] * wk * pair_est;
	    if ((use_ub) && (fs_k + bs_k <= wk) && (est_cost > wk * ubound)){
	      goto fs_next_edge;
	    }
	    edge_cost += pair_edge_cost + weights[j][k] * fs_k * pair_est;
	    if(dir_mask & BMASK){
	      edge_cost
		-= weights[j][k] * bs_k *
		      (lookup_estimate(bs_estimator[sub],
				       next_pos[j], next_pos[k])
		       + add_cost * (next_pos[j] + next_pos[k]));
	    }
	    sub++;
	  }
	}
	if((use_ub) && (est_cost > wk * ubound)){
	  goto fs_next_edge;
	}
	if((next_node = get_node(node_set, next_pos)) == NULL){
	  next_node = create_node(node_set, next_pos);
	}
	if(fs_info(next_node) == NULL){
	  fs_info_init(next_node);
	}
	if(is_bs_finished(next_node)){
	  is_bridge = 1;
	}
#ifdef VERBOSE
	if(lookup_fs_status(next_node) == UNVISITED){
	  fs_visited++;
	  if((fs_visited & 0x7fff) == 0){
	    fprintf(stderr,
		    ".... visited %d nodes (forward search)\n", fs_visited);
	  }
	}
#endif
	extend_node(fs_pqueue, node, next_node, edge_cost);
/*	printf("%d(%d,%d,%d) ", lookup_cost(node, FORWARD) + edge_cost - lookup_cost(next_node, FORWARD),lookup_cost(node, FORWARD),lookup_cost(next_node, FORWARD), edge_cost);*/
      fs_next_edge:
      }
      if((node == dest_node) || is_bs_finished(node)){
	mid_node = node;
	break;
      }
      if(is_bridge){
	bridge_node_list = node_list_cons(node, bridge_node_list);
      }
    }
    if(dir_mask & BMASK){
      if((node = get_node_from_pqueue(bs_pqueue)) == NULL){
	fprintf(stderr, "Upper bound is too small\n");
	exit(-1);
      }
      set_bs_status(node, FINISHED);
      bs_farthest = lookup_bs_cost(node);
#ifdef VERBOSE
      bs_expanded++;
#endif
      for(i = 1; i < (1 << dim); i++){
	n = 1;
	for(j = 0; j < dim; j++){
	  if(i & n){
	    if((next_pos[j] = node->pos[j] - 1) < 0){
	      goto bs_next_edge;
	    }
	  }else{
	    next_pos[j] = node->pos[j];
	  }
	  n <<= 1;
	}
	edge_cost = 0;
	est_cost = lookup_bs_cost(node);
	sub = 0;
	for(j = 0; j < dim; j++){
	  for(k = j + 1; k < dim; k ++){
	    pair_edge_cost = 0;
	    if((node->pos[j] != next_pos[j]) && (node->pos[k] != next_pos[k])){
	      for(l = 0; l < alignment[j].dim; l++){
		for(m = 0; m < alignment[k].dim; m++){
		  pair_edge_cost += weights[j][k] * wk
		    * matching_cost(alignment[j].seq[l].str[next_pos[j]],
				    alignment[k].seq[m].str[next_pos[k]]);
		}
	      }
	      pair_edge_cost += weights[j][k] * wk * add_cost * 2;
	    }else if(node->pos[j] != next_pos[j]){
	      n = 0; 
	      for(l = 0; l < alignment[j].dim; l++){
		if(is_normal_char(alignment[j].seq[l].str[next_pos[j]])){
		  n++;
		}
	      }
	      for(m = 0; m < alignment[k].dim; m++){
		if(((next_pos[k] == 0)
		    || (alignment[k].seq[m].str[next_pos[k] - 1]
			== EXTERNAL_GAP_CHAR))
		   || ((next_pos[k] == alignment[k].seq[0].len)
		       || (alignment[k].seq[m].str[next_pos[k]]
			   == EXTERNAL_GAP_CHAR))){
		  pair_edge_cost += weights[j][k] * wk * n * (-external_gap_penalty);
		}else{
		  pair_edge_cost += weights[j][k] * wk * n * (-internal_gap_penalty);
		}
	      }
	      pair_edge_cost += weights[j][k] * wk * add_cost;
	    }else if(node->pos[k] != next_pos[k]){
	      n = 0;
	      for(m = 0; m < alignment[k].dim; m++){
		if(is_normal_char(alignment[k].seq[m].str[next_pos[k]])){
		  n++;
		}
	      }
	      for(l = 0; l < alignment[j].dim; l++){
		if(((next_pos[j] == 0)
		    || (alignment[j].seq[l].str[next_pos[j] - 1]
			== EXTERNAL_GAP_CHAR))
		   || ((next_pos[j] == alignment[j].seq[0].len)
		       || (alignment[j].seq[l].str[next_pos[j]]
			   == EXTERNAL_GAP_CHAR))){
		  pair_edge_cost += weights[j][k] * wk * n * (-external_gap_penalty);
		}else{
		  pair_edge_cost += weights[j][k] * wk * n * (-internal_gap_penalty);
		}
	      }
	      pair_edge_cost += weights[j][k] * wk * add_cost;
	    }
	    if(dir_mask & FMASK){
	      pair_edge_cost
		+= weights[j][k] * fs_k *
		      (lookup_estimate(fs_estimator[sub],
				       node->pos[j], node->pos[k])
		       + add_cost * ((alignment[j].seq[0].len
				      - node->pos[j])
				     + (alignment[k].seq[0].len
					- node->pos[k])));
	    }
	    pair_edge_cost
	      -= weights[j][k] * bs_k *
		   (lookup_estimate(bs_estimator[sub],
				    node->pos[j], node->pos[k])
		    + add_cost * (node->pos[j] + node->pos[k]));
	    pair_est
	      = lookup_estimate(bs_estimator[sub],
				next_pos[j], next_pos[k])
		+ add_cost * (next_pos[j] + next_pos[k]);
	    est_cost += pair_edge_cost + weights[j][k] * wk * pair_est;
	    if((use_ub) && (fs_k + bs_k <= wk) && (est_cost > wk * ubound)){
	      goto bs_next_edge;
	    }
	    edge_cost += pair_edge_cost + weights[j][k] * bs_k * pair_est;
	    if(dir_mask & FMASK){
	      edge_cost
		-= weights[j][k] * fs_k *
		     (lookup_estimate(fs_estimator[sub],
				      next_pos[j], next_pos[k])
		      + add_cost * ((alignment[j].seq[0].len
				     - next_pos[j])
				    + (alignment[k].seq[0].len
				       - next_pos[k])));
	    }
	    sub++;
	  }
	}
	if((use_ub) && (est_cost > wk * ubound)){
	  goto bs_next_edge;
	}
	if((next_node = get_node(node_set, next_pos)) == NULL){
	  next_node = create_node(node_set, next_pos);
	}
	if(bs_info(next_node) == NULL){
	  bs_info_init(next_node);
	}
	if(is_fs_finished(next_node)){
	  bridge_node_list = node_list_cons(next_node, bridge_node_list);
	}
#ifdef VERBOSE
	if(lookup_bs_status(next_node) == UNVISITED){
	  bs_visited++;
	  if((bs_visited & 0x7fff) == 0){
	    fprintf(stderr,
		    ".... visited %d nodes (backward search)\n", bs_visited);
	  }
	}
#endif
	extend_node(bs_pqueue, node, next_node, edge_cost);
      bs_next_edge:
      }
      if((node == orig_node) || is_fs_finished(node)){
	mid_node = node;
	break;
      }
    }
  }

  shortest_cost = lookup_fs_cost(mid_node) + lookup_bs_cost(mid_node);
  for(node_list = bridge_node_list;
      node_list != NULL; node_list = node_list->next){
    if((tmp = lookup_fs_cost(node_list->node) + lookup_bs_cost(node_list->node))
       < shortest_cost){
      mid_node = node_list->node;
      shortest_cost = tmp;
    }
  }

  if ((use_ub)&&((ubound - delta) * wk < shortest_cost)) {
    delta = (ubound - shortest_cost) / wk;
    fprintf(stderr, "Upper bound is too small:\n");
    fprintf(stderr, "- Updated delta to %d",delta);
    if (gd) fprintf(stderr, "/%d", gap_denominator);
    fprintf(stderr, "\n");
  }

  new_ubound = shortest_cost + delta*wk;

#ifdef DEBUG
  fprintf(stdout,"new ubound = %d",new_ubound);
  if (gd) fprintf(stdout, "/%d", gap_denominator);
  fprintf(stdout, "\n");
  fprintf(stdout,"shortest cost = %d",shortest_cost);
  if (gd) fprintf(stdout, "/%d", gap_denominator);
  fprintf(stdout, "\n");
  fprintf(stdout,"for/back_ward largest cost = %d/%d",fs_farthest,bs_farthest);
  if (gd) fprintf(stdout, " (/%d)", gap_denominator);
  fprintf(stdout, "\n");
#endif

#ifdef VERBOSE
  times(&t_optimal_end);
#endif

#ifdef VERBOSE
  fs_visited_for_opt = fs_visited;
  bs_visited_for_opt = bs_visited;
  fs_expanded_for_opt = fs_expanded;
  bs_expanded_for_opt = bs_expanded;
#endif

  while(1){
    if(dir_mask & FMASK){
      if((node = get_node_from_pqueue(fs_pqueue)) == NULL) break;
      set_fs_status(node, FINISHED);
      fs_farthest=lookup_fs_cost(node);
#ifdef VERBOSE
      fs_expanded++;
#endif
      if(fs_farthest+bs_farthest > new_ubound) break;
      is_bridge = 0;
      for(i = 1; i < (1 << dim); i++){
	n = 1;
	for(j = 0; j < dim; j++){
	  if(i & n){
	    if((next_pos[j] = node->pos[j] + 1) > alignment[j].seq[0].len){
	      goto fs_next_edge2;
	    }
	  }else{
	    next_pos[j] = node->pos[j];
	  }
	  n <<= 1;
	}
	edge_cost = 0;
	est_cost = lookup_fs_cost(node);
	sub = 0;
	for(j = 0; j < dim; j++){
	  for(k = j + 1; k < dim; k ++){
	    pair_edge_cost = 0;
	    if((node->pos[j] != next_pos[j]) && (node->pos[k] != next_pos[k])){
	      for(l = 0; l < alignment[j].dim; l++){
		for(m = 0; m < alignment[k].dim; m++){
		  pair_edge_cost += weights[j][k] * wk
		    * matching_cost(alignment[j].seq[l].str[node->pos[j]],
				    alignment[k].seq[m].str[node->pos[k]]);
		}
	      }
	      pair_edge_cost += weights[j][k] * wk * add_cost * 2;
	    }else if(node->pos[j] != next_pos[j]){
	      n = 0; 
	      for(l = 0; l < alignment[j].dim; l++){
		if(is_normal_char(alignment[j].seq[l].str[node->pos[j]])){
		  n++;
		}
	      }
	      for(m = 0; m < alignment[k].dim; m++){
		if(((node->pos[k] == 0)
		    || (alignment[k].seq[m].str[node->pos[k] - 1]
			== EXTERNAL_GAP_CHAR))
		   || ((node->pos[k] == alignment[k].seq[0].len)
		       || (alignment[k].seq[m].str[node->pos[k]]
			   == EXTERNAL_GAP_CHAR))){
		  pair_edge_cost += weights[j][k] * wk * n * (-external_gap_penalty);
		}else{
		  pair_edge_cost += weights[j][k] * wk * n * (-internal_gap_penalty);
		}
	      }
	      pair_edge_cost += weights[j][k] * wk * add_cost;
	    }else if(node->pos[k] != next_pos[k]){
	      n = 0;
	      for(m = 0; m < alignment[k].dim; m++){
		if(is_normal_char(alignment[k].seq[m].str[node->pos[k]])){
		  n++;
		}
	      }
	      for(l = 0; l < alignment[j].dim; l++){
		if(((node->pos[j] == 0)
		    || (alignment[j].seq[l].str[node->pos[j] - 1]
			== EXTERNAL_GAP_CHAR))
		   || ((node->pos[j] == alignment[j].seq[0].len)
		       || (alignment[j].seq[l].str[node->pos[j]]
			   == EXTERNAL_GAP_CHAR))){
		  pair_edge_cost += weights[j][k] * wk * n * (-external_gap_penalty);
		}else{
		  pair_edge_cost += weights[j][k] * wk * n * (-internal_gap_penalty);
		}
	      }
	      pair_edge_cost += weights[j][k] * wk * add_cost;
	    }
	    pair_edge_cost
	      -= fs_k * weights[j][k] *
		        (lookup_estimate(fs_estimator[sub],
					 node->pos[j], node->pos[k])
			 + add_cost * ((alignment[j].seq[0].len
					- node->pos[j])
				       + (alignment[k].seq[0].len
					  - node->pos[k])));
	    if(dir_mask & BMASK){
	      pair_edge_cost
		+= weights[j][k] * bs_k *
		       (lookup_estimate(bs_estimator[sub],
					node->pos[j], node->pos[k])
			+ add_cost * (node->pos[j] + node->pos[k]));
	    }
	    pair_est
	      = lookup_estimate(fs_estimator[sub],
				next_pos[j], next_pos[k])
	      + add_cost * ((alignment[j].seq[0].len - next_pos[j])
			    + (alignment[k].seq[0].len - next_pos[k]));
	    est_cost += pair_edge_cost + weights[j][k] * wk * pair_est;
	    if (est_cost > new_ubound) {
	      goto fs_next_edge2;
	    }
	    edge_cost += pair_edge_cost + weights[j][k] * fs_k * pair_est;
	    if(dir_mask & BMASK){
	      edge_cost
		-= weights[j][k] * bs_k *
		      (lookup_estimate(bs_estimator[sub],
				       next_pos[j], next_pos[k])
		       + add_cost * (next_pos[j] + next_pos[k]));
	    }
	    sub++;
	  }
	}
	if(est_cost > new_ubound) {
	  goto fs_next_edge2;
	}
	if((next_node = get_node(node_set, next_pos)) == NULL){
	  next_node = create_node(node_set, next_pos);
	}
	if(fs_info(next_node) == NULL){
	  fs_info_init(next_node);
	}
	if(is_bs_finished(next_node)){
	  is_bridge = 1;
	}
#ifdef VERBOSE
	if(lookup_fs_status(next_node) == UNVISITED){
	  fs_visited++;
	  if((fs_visited & 0x7fff) == 0){
	    fprintf(stderr,
		    ".... visited %d nodes (forward search)\n", fs_visited);
	  }
	}
#endif
	extend_node(fs_pqueue, node, next_node, edge_cost);
      fs_next_edge2:
      }
      if(is_bridge){
	bridge_node_list = node_list_cons(node, bridge_node_list);
      }
    }
    if(dir_mask & BMASK){
      if((node = get_node_from_pqueue(bs_pqueue)) == NULL){
	break;
      }
      set_bs_status(node, FINISHED);
      bs_farthest = lookup_bs_cost(node);
#ifdef VERBOSE
      bs_expanded++;
#endif
      if(fs_farthest+bs_farthest > new_ubound){
	break;
      }
      for(i = 1; i < (1 << dim); i++){
	n = 1;
	for(j = 0; j < dim; j++){
	  if(i & n){
	    if((next_pos[j] = node->pos[j] - 1) < 0){
	      goto bs_next_edge2;
	    }
	  }else{
	    next_pos[j] = node->pos[j];
	  }
	  n <<= 1;
	}
	edge_cost = 0;
	est_cost = lookup_bs_cost(node);
	sub = 0;
	for(j = 0; j < dim; j++){
	  for(k = j + 1; k < dim; k ++){
	    pair_edge_cost = 0;
	    if((node->pos[j] != next_pos[j]) && (node->pos[k] != next_pos[k])){
	      for(l = 0; l < alignment[j].dim; l++){
		for(m = 0; m < alignment[k].dim; m++){
		  pair_edge_cost += weights[j][k] * wk
		    * matching_cost(alignment[j].seq[l].str[next_pos[j]],
				    alignment[k].seq[m].str[next_pos[k]]);
		}
	      }
	      pair_edge_cost += weights[j][k] * wk * add_cost * 2;
	    }else if(node->pos[j] != next_pos[j]){
	      n = 0; 
	      for(l = 0; l < alignment[j].dim; l++){
		if(is_normal_char(alignment[j].seq[l].str[next_pos[j]])){
		  n++;
		}
	      }
	      for(m = 0; m < alignment[k].dim; m++){
		if(((next_pos[k] == 0)
		    || (alignment[k].seq[m].str[next_pos[k] - 1]
			== EXTERNAL_GAP_CHAR))
		   || ((next_pos[k] == alignment[k].seq[0].len)
		       || (alignment[k].seq[m].str[next_pos[k]]
			   == EXTERNAL_GAP_CHAR))){
		  pair_edge_cost += weights[j][k] * wk * n * (-external_gap_penalty);
		}else{
		  pair_edge_cost += weights[j][k] * wk * n * (-internal_gap_penalty);
		}
	      }
	      pair_edge_cost += weights[j][k] * wk * add_cost;
	    }else if(node->pos[k] != next_pos[k]){
	      n = 0;
	      for(m = 0; m < alignment[k].dim; m++){
		if(is_normal_char(alignment[k].seq[m].str[next_pos[k]])){
		  n++;
		}
	      }
	      for(l = 0; l < alignment[j].dim; l++){
		if(((next_pos[j] == 0)
		    || (alignment[j].seq[l].str[next_pos[j] - 1]
			== EXTERNAL_GAP_CHAR))
		   || ((next_pos[j] == alignment[j].seq[0].len)
		       || (alignment[j].seq[l].str[next_pos[j]]
			   == EXTERNAL_GAP_CHAR))){
		  pair_edge_cost += weights[j][k] * wk * n * (-external_gap_penalty);
		}else{
		  pair_edge_cost += weights[j][k] * wk * n * (-internal_gap_penalty);
		}
	      }
	      pair_edge_cost += weights[j][k] * wk * add_cost;
	    }
	    if(dir_mask & FMASK){
	      pair_edge_cost
		+= weights[j][k] * fs_k *
		          (lookup_estimate(fs_estimator[sub],
					   node->pos[j], node->pos[k])
			   + add_cost * ((alignment[j].seq[0].len
					  - node->pos[j])
					 + (alignment[k].seq[0].len
					    - node->pos[k])));
	    }
	    pair_edge_cost
	      -= bs_k * weights[j][k] *
		        (lookup_estimate(bs_estimator[sub],
					 node->pos[j], node->pos[k])
			 + add_cost * (node->pos[j] + node->pos[k]));
	    pair_est
	      = lookup_estimate(bs_estimator[sub],
				next_pos[j], next_pos[k])
		+ add_cost * (next_pos[j] + next_pos[k]);
	    est_cost += pair_edge_cost + weights[j][k] * wk * pair_est;
	    if(est_cost > new_ubound){
	      goto bs_next_edge2;
	    }
	    edge_cost += pair_edge_cost + bs_k * pair_est;
	    if(dir_mask & FMASK){
	      edge_cost
		-= weights[j][k] * fs_k *
		          (lookup_estimate(fs_estimator[sub],
					   next_pos[j], next_pos[k])
			   + add_cost * ((alignment[j].seq[0].len
					  - next_pos[j])
					 + (alignment[k].seq[0].len
					    - next_pos[k])));
	    }
	    sub++;
	  }
	}
	if(est_cost > new_ubound){
	  goto bs_next_edge2;
	}
	if((next_node = get_node(node_set, next_pos)) == NULL){
	  next_node = create_node(node_set, next_pos);
	}
	if(bs_info(next_node) == NULL){
	  bs_info_init(next_node);
	}
	if(is_fs_finished(next_node)){
	  bridge_node_list = node_list_cons(next_node, bridge_node_list);
	}
#ifdef VERBOSE
	if(lookup_bs_status(next_node) == UNVISITED){
	  bs_visited++;
	  if((bs_visited & 0x7fff) == 0){
	    fprintf(stderr,
		    ".... visited %d nodes (backward search)\n", bs_visited);
	  }
	}
#endif
	extend_node(bs_pqueue, node, next_node, edge_cost);
      bs_next_edge2:
      }
    }
  }

  if(dir_mask & FMASK) {
    set_status(orig_node, FORWARD, BESTPATH);
  }
  if(dir_mask & BMASK) {
    set_status(dest_node, BACKWARD, BESTPATH);
  }

#ifdef VERBOSE
  times(&t_search_end);
  fprintf(stderr, "Finished searching with A* Algorithm\n");
#endif

/* From here, Eppstein's algorithm begins! */

  if ((print_subopt)||(opt_sols!=NULL)||(opt_sols2!=NULL)||(opt_sols3!=NULL)) {
    maxdepth=alignment[0].seq[0].len;
    if (maxdepth < 16) maxdepth = 16;
    init_subopt_answer(answer, maxdepth);
    depth = 0;        /* No sidetrack = optimal */
  }

  if(opt_sols != NULL) {
    opt_sols->number_of_sols = 0;
    opt_sols->number_of_max_sols = 0;
    opt_sols->number_of_min_sols = 0;
  }
  if(opt_sols2 != NULL) {
    opt_sols2->number_of_sols = 0;
    opt_sols2->number_of_max_sols = 0;
    opt_sols2->number_of_min_sols = 0;
  }
  if(opt_sols3 != NULL) {
    opt_sols3->number_of_sols = 0;
    opt_sols3->number_of_max_sols = 0;
    opt_sols3->number_of_min_sols = 0;
  }

  if(kbs <= 0) {
/* make histogram */

    if((histogram = (int *)malloc((delta+1)*sizeof(int))) == NULL) {
      fprintf(stderr,"Can't allocate memory for the histogram\n");
      exit(-1);
    }

    histogram[0] = 1; /* optimal solution */
    for(i=1;i<=delta;i++) histogram[i] = 0;

 /* Notice that bidirectional is not implemented ! */
    if(dir_mask & FMASK){
      construct_heap(mid_node, node_set, dim, alignment, 
		     wk, fs_k, bs_k, FORWARD, delta,
#ifdef VERBOSE
		     constructed, h_t, h_edge,
#endif
		     add_cost, fs_estimator, BESTPATH);
      if(use_detour) {
	make_histogram_for_h_t_d1(lookup_tree_heap(mid_node, FORWARD), 0, histogram,
				  mid_node, orig_node, dest_node,
				  node_set, dim, alignment, 
				  wk, fs_k, bs_k, FORWARD, delta,
#ifdef VERBOSE
				  constructed, h_t, h_edge,
#endif
				  add_cost, fs_estimator,
				  print_subopt, answer, depth, &maxdepth, line);
      } else {
	make_histogram_for_h_t(lookup_tree_heap(mid_node, FORWARD), 0, histogram,
			       mid_node,  orig_node, dest_node,
			       node_set, dim, alignment, 
			       wk, fs_k, bs_k, FORWARD, delta,
#ifdef VERBOSE
			       constructed, h_t, h_edge,
#endif
			       add_cost, fs_estimator,
			       print_subopt, answer, depth, &maxdepth, line,
			       opt_sols, opt_sols2, opt_sols3);
      }
    } else {
      construct_heap(mid_node, node_set, dim, alignment, 
		     wk, fs_k, bs_k, BACKWARD, delta,
#ifdef VERBOSE
		     constructed, h_t, h_edge,
#endif
		     add_cost, bs_estimator, BESTPATH);
      if(use_detour) {
	make_histogram_for_h_t_d1(lookup_tree_heap(mid_node, BACKWARD), 0, histogram,
				  mid_node,  orig_node, dest_node,
				  node_set, dim, alignment, 
				  wk, fs_k, bs_k, BACKWARD, delta,
#ifdef VERBOSE
				  constructed, h_t, h_edge,
#endif
				  add_cost, bs_estimator,
				  print_subopt, answer, depth, &maxdepth, line);
      } else {
	make_histogram_for_h_t(lookup_tree_heap(mid_node, BACKWARD), 0, histogram,
			       mid_node,  orig_node, dest_node,
			       node_set, dim, alignment, 
			       wk, fs_k, bs_k, BACKWARD, delta,
#ifdef VERBOSE
			       constructed, h_t, h_edge,
#endif
			       add_cost, bs_estimator,
			       print_subopt, answer, depth, &maxdepth, line,
			       opt_sols, opt_sols2, opt_sols3);
      }
    }

#ifdef VERBOSE
    total_sols = 0;
#endif

    printf(" Delta   #solutions   #total\n");
    for(i=0;i<=delta;i++) {
      total_sols +=histogram[i];
      printf("%5d",i);
      if (gd) printf("/%d (=%6.3f)", gap_denominator, (((float)i)/((float)gap_denominator)));
      printf(":   %8d  %8d\n", histogram[i], total_sols);
    }
  } else {
/* solve k best solutions */
/* Not implimented yet...^^; */
  }

#ifdef VERBOSE
  times(&t_eppstein_end);
#endif

  if(dir_mask & FMASK){
    sub = 0;
    for(i = 0; i < dim; i++){
      for(j = i + 1; j < dim; j++){
	estimator_finish(fs_estimator[sub]);
	sub++;
      }
    }
    pqueue_finish(fs_pqueue);
  }
  if(dir_mask & BMASK){
    sub = 0;
    for(i = 0; i < dim; i++){
      for(j = i + 1; j < dim; j++){
	estimator_finish(bs_estimator[sub]);
	sub++;
      }
    }
    pqueue_finish(bs_pqueue);
  }


  if((al_p = (alignment_t *)malloc(sizeof(alignment_t))) == NULL){
    fprintf(stderr, "Can't allocate memory for result alignment");
    exit(-1);
  }
  al_p->dim = 0;
  for(i = 0; i < dim; i++){
    al_p->dim += alignment[i].dim;
  }
  al_len = 0;
  for(node = mid_node; node != orig_node; node = lookup_fs_prev_node(node)){
    al_len++;
  }
  for(node = mid_node; node != dest_node; node = lookup_bs_prev_node(node)){
    al_len++;
  }
  if((al_p->seq = (sequence_t *)malloc(sizeof(sequence_t) * al_p->dim))
     == NULL){
    fprintf(stderr, "Can't allocate memory for result alignment");
    exit(-1);
  }
  k = 0;
  for(i = 0; i < dim; i++){
    for(j = 0; j < alignment[i].dim; j++){
      al_p->seq[k].len = al_len;
      if((al_p->seq[k].str
	  = (char *)malloc(sizeof(char) * (al_len + 1))) == NULL){
	fprintf(stderr, "Can't allocate memory for result alignment");
	exit(-1);
      }
      al_p->seq[k].str[al_len] = '\0';
      p = -1;
      for(node = mid_node; node != orig_node;
	  node = lookup_fs_prev_node(node)){
	p++;
      }
      q = mid_node->pos[i] - 1;
      for(node = mid_node; node != orig_node;
	  node = lookup_fs_prev_node(node)){
	if(node->pos[i] != lookup_fs_prev_node(node)->pos[i]){
	  al_p->seq[k].str[p--] = alignment[i].seq[j].str[q--];
	}else{
	  if(((q + 1 > alignment[i].seq[0].len - 1)
	      || (alignment[i].seq[j].str[q + 1] == EXTERNAL_GAP_CHAR))
	     || ((q < 0)
		 || (alignment[i].seq[j].str[q] == EXTERNAL_GAP_CHAR))){
	    al_p->seq[k].str[p--] = EXTERNAL_GAP_CHAR;
	  }else{
	    al_p->seq[k].str[p--] = INTERNAL_GAP_CHAR;
	  }
	}
      }
      p = al_len;
      for(node = mid_node; node != dest_node;
	  node = lookup_bs_prev_node(node)){
	p--;
      }
      q = mid_node->pos[i];
      for(node = mid_node; node != dest_node;
	  node = lookup_bs_prev_node(node)){
	if(node->pos[i] != lookup_bs_prev_node(node)->pos[i]){
	  al_p->seq[k].str[p++] = alignment[i].seq[j].str[q++];
	}else{
	  if(((q > alignment[i].seq[0].len - 1)
	      || (alignment[i].seq[j].str[q] == EXTERNAL_GAP_CHAR))
	     || ((q - 1 < 0)
		 || (alignment[i].seq[j].str[q - 1] == EXTERNAL_GAP_CHAR))){
	    al_p->seq[k].str[p++] = EXTERNAL_GAP_CHAR;
	  }else{
	    al_p->seq[k].str[p++] = INTERNAL_GAP_CHAR;
	  }
	}
      }
      k++;
    }
  }
  al_score = -((lookup_fs_cost(mid_node) + lookup_bs_cost(mid_node)) / wk);

  if (((opt_sols == NULL)&&(opt_sols2 == NULL)&&(opt_sols3 == NULL))
      || (print_subopt)) {
    printf("Optimal alignment:\n");
    print_alignment(al_p, line);
    printf("\n");
    print_number_of_gaps(al_p, (&ig), (&eg), (&is), (&es), 1);
  } else print_number_of_gaps(al_p, (&ig), (&eg), (&is), (&es), 0);

#ifdef VERBOSE
  fprintf(stderr, "Estimated inter-group cost: %d",
	  (lookup_fs_cost(orig_node) + lookup_bs_cost(dest_node)) / wk);
  if (gd) fprintf(stderr, "/%d", gap_denominator);
  fprintf(stderr, "\n");
  fprintf(stderr, "Actual inter-group cost: %d\n", -al_score);
  if (gd) fprintf(stderr, "/%d", gap_denominator);
  fprintf(stderr, "\n");
#endif

  for(i = 0; i < dim; i++){
    for(j = i + 1; j < dim; j++){
      al_score
	+= add_cost * (alignment[i].seq[0].len + alignment[j].seq[0].len);
    }
  }
  al_p->score = al_score - intra_group_cost;
  printf("Optimal score = %d", al_p->score);
  if (gd) printf("/%d", gap_denominator);
  printf("\n\n");
/*  printf("Score: %d\n", weighted_score(al_p, weights));*/

  if (opt_sols2 != NULL) {
    (opt_sols2->number_of_sols)++;
    opt_sols2->score = al_p->score;
    slope = weighted_score(al_p, weights3);
    if (opt_sols2->max_alignment == NULL) {
      opt_sols2->max_alignment = al_p;
      opt_sols2->min_alignment = al_p;
      opt_sols2->max_slope = slope;
      opt_sols2->min_slope = slope;
      opt_sols2->number_of_max_sols = 1;
      opt_sols2->number_of_min_sols = 1;
      if (print_subopt) {
	printf("This is temporally both min and max\n======\n");
      }
    } else {
      s = 1;
      if (opt_sols2->max_slope == slope) {
	(opt_sols2->number_of_max_sols)++;
	if (print_subopt) {
	  printf("This is temporally max\n======\n");
	}
      } else if (opt_sols2->max_slope < slope) {
	opt_sols2->max_slope = slope;
	opt_sols2->number_of_max_sols = 1;
	if (opt_sols2->max_alignment != opt_sols2->min_alignment) {
	  free_alignment(opt_sols2->max_alignment, 1);
	}
	s = 0;
	opt_sols2->max_alignment = al_p;
      }

      if (opt_sols2->min_slope == slope) {
	(opt_sols2->number_of_min_sols)++;
	if (print_subopt) {
	  printf("This is temporally min\n======\n");
	}
      } else if (opt_sols2->min_slope > slope) {
	opt_sols2->min_slope = slope;
	opt_sols2->number_of_min_sols = 1;
	if (opt_sols2->max_alignment != opt_sols2->min_alignment) {
	  free_alignment(opt_sols2->min_alignment, 1);
	}
	s = 0;
	opt_sols2->min_alignment = al_p;
      }
      if (s) free_alignment(al_p, 1);
    }
  } else if (opt_sols3 != NULL) {
    (opt_sols3->number_of_sols)++;
    opt_sols3->score = al_p->score;
    slope = score_slope(al_p);
    if (opt_sols3->max_alignment == NULL) {
      opt_sols3->max_alignment = al_p;
      opt_sols3->min_alignment = al_p;
      opt_sols3->max_slope = slope;
      opt_sols3->min_slope = slope;
      opt_sols3->number_of_max_sols = 1;
      opt_sols3->number_of_min_sols = 1;
      if (print_subopt) {
	printf("This is temporally both min and max\n======\n");
      }
    } else {
      s = 1;
      if (opt_sols3->max_slope == slope) {
	(opt_sols3->number_of_max_sols)++;
	if (print_subopt) {
	  printf("This is temporally max\n======\n");
	}
      } else if (opt_sols3->max_slope < slope) {
	opt_sols3->max_slope = slope;
	opt_sols3->number_of_max_sols = 1;
	if (opt_sols3->max_alignment != opt_sols3->min_alignment) {
	  free_alignment(opt_sols3->max_alignment, 1);
	}
	s = 0;
	opt_sols3->max_alignment = al_p;
      }

      if (opt_sols3->min_slope == slope) {
	(opt_sols3->number_of_min_sols)++;
	if (print_subopt) {
	  printf("This is temporally min\n======\n");
	}
      } else if (opt_sols3->min_slope > slope) {
	opt_sols3->min_slope = slope;
	opt_sols3->number_of_min_sols = 1;
	if (opt_sols3->max_alignment != opt_sols3->min_alignment) {
	  free_alignment(opt_sols3->min_alignment, 1);
	}
	s = 0;
	opt_sols3->min_alignment = al_p;
      }
      if (s) free_alignment(al_p, 1);
    }
  } else if (opt_sols != NULL) {
    opt_sols->gap_penalty = internal_gap_penalty;
    opt_sols->gap_denominator = gap_denominator;
    opt_sols->score = al_p->score;
    tg = ig + eg;
    (opt_sols->number_of_sols)++;
    if (opt_sols->max_alignment == NULL) {
      opt_sols->max_alignment = al_p;
      opt_sols->min_alignment = al_p;
      opt_sols->max_gap = tg;
      opt_sols->min_gap = tg;
      opt_sols->number_of_max_sols = 1;
      opt_sols->number_of_min_sols = 1;
      if (print_subopt) {
	printf("This is temporally both min and max\n======\n");
      }
    } else {
      s = 1;
      if (opt_sols->max_gap == tg) {
	(opt_sols->number_of_max_sols)++;
	if (print_subopt) {
	  printf("This is temporally max\n======\n");
	}
      } else if (opt_sols->max_gap < tg) {
	opt_sols->max_gap = tg;
	opt_sols->number_of_max_sols = 1;
	if (opt_sols->max_alignment != opt_sols->min_alignment) {
	  free_alignment(opt_sols->max_alignment, 1);
	}
	s = 0;
	opt_sols->max_alignment = al_p;
      }

      if (opt_sols->min_gap == tg) {
	(opt_sols->number_of_min_sols)++;
	if (print_subopt) {
	  printf("This is temporally min\n======\n");
	}
      } else if (opt_sols->min_gap > tg) {
	opt_sols->min_gap = tg;
	opt_sols->number_of_min_sols = 1;
	if (opt_sols->max_alignment != opt_sols->min_alignment) {
	  free_alignment(opt_sols->min_alignment, 1);
	}
	s = 0;
	opt_sols->min_alignment = al_p;
      }
      if (s) free_alignment(al_p, 1);
    }
  } else {
    free_alignment(al_p, 1);
  }

#ifdef VERBOSE
  times(&t_end);
  wnodes = alignment[0].seq[0].len + 1;
  for(i = 1; i < dim; i++){
    wnodes *= alignment[i].seq[0].len + 1;
  }

  fprintf(stderr, "# of whole nodes: %8.2e\n", wnodes);

  if (fs_visited != 0) fprintf(stderr, "# of visited nodes (forward)              : %d\n", fs_visited);
  if (fs_expanded != 0) fprintf(stderr, "# of expanded nodes (forward)             : %d\n", fs_expanded);
  if (bs_visited != 0) fprintf(stderr, "# of visited nodes (backward)             : %d\n", bs_visited);
  if (bs_expanded != 0) fprintf(stderr, "# of expanded nodes (backward)            : %d\n", bs_expanded);

  if (fs_visited_for_opt != 0) fprintf(stderr, "# of visited nodes for opt (forward)      : %d\n", fs_visited_for_opt);
  if (fs_expanded_for_opt != 0) fprintf(stderr, "# of expanded nodes for opt (forward)     : %d\n", fs_expanded_for_opt);
  if (bs_visited_for_opt != 0) fprintf(stderr, "# of visited nodes for opt (backward)     : %d\n", bs_visited_for_opt);
  if (bs_expanded_for_opt != 0) fprintf(stderr, "# of expanded nodes for opt (backward)    : %d\n", bs_expanded_for_opt);

  if (constructed[FORWARD] != 0) fprintf(stderr, "# of nodes with Eppstein's heap (forward) : %d\n", constructed[FORWARD]);
  if (h_t[FORWARD] != 0) fprintf(stderr, "# of nodes in H_T (forward)               : %d\n", h_t[FORWARD]);
  if (h_edge[FORWARD] != 0) fprintf(stderr, "# of nodes in H_edge (forward)            : %d\n", h_edge[FORWARD]);

  if (constructed[BACKWARD] != 0) fprintf(stderr, "# of nodes with Eppstein's heap(backward) : %d\n", constructed[BACKWARD]);
  if (h_t[BACKWARD] != 0) fprintf(stderr, "# of nodes in H_T (backward)              : %d\n", h_t[BACKWARD]);
  if (h_edge[BACKWARD] != 0) fprintf(stderr, "# of nodes in H_edge (backward)           : %d\n", h_edge[BACKWARD]);

/*  if (kbs<=0)*/ fprintf(stderr, "# of solutions examined                   : %d\n", total_sols);

  fprintf(stderr, "Used time:          User    System  Total\n");
#if defined(linux)
  fprintf(stderr, "  Pre-process      %6.2f  %6.2f  %6.2f\n",
	  (float)(t_search_start.tms_utime - t_start.tms_utime) / CLK_TCK,
	  (float)(t_search_start.tms_stime - t_start.tms_stime) / CLK_TCK,
	  (float)((t_search_start.tms_utime - t_start.tms_utime)
		  + (t_search_start.tms_stime - t_start.tms_stime)) / CLK_TCK);
  fprintf(stderr, "  Search(opt)      %6.2f  %6.2f  %6.2f\n",
	  (float)(t_optimal_end.tms_utime - t_search_start.tms_utime) / CLK_TCK,
	  (float)(t_optimal_end.tms_stime - t_search_start.tms_stime) / CLK_TCK,
	  (float)((t_optimal_end.tms_utime - t_search_start.tms_utime)
		  + (t_optimal_end.tms_stime - t_search_start.tms_stime))
	  / CLK_TCK);
  fprintf(stderr, "  Search(total)    %6.2f  %6.2f  %6.2f\n",
	  (float)(t_search_end.tms_utime - t_search_start.tms_utime) / CLK_TCK,
	  (float)(t_search_end.tms_stime - t_search_start.tms_stime) / CLK_TCK,
	  (float)((t_search_end.tms_utime - t_search_start.tms_utime)
		  + (t_search_end.tms_stime - t_search_start.tms_stime))
	  / CLK_TCK);
  fprintf(stderr, "  Eppstein Process %6.2f  %6.2f  %6.2f\n",
	  (float)(t_eppstein_end.tms_utime - t_search_end.tms_utime) / CLK_TCK,
	  (float)(t_eppstein_end.tms_stime - t_search_end.tms_stime) / CLK_TCK,
	  (float)((t_eppstein_end.tms_utime - t_search_end.tms_utime)
		  + (t_eppstein_end.tms_stime - t_search_end.tms_stime))
	  / CLK_TCK);
  fprintf(stderr, "  Post-process     %6.2f  %6.2f  %6.2f\n",
	  (float)(t_end.tms_utime - t_eppstein_end.tms_utime) / CLK_TCK,
	  (float)(t_end.tms_stime - t_eppstein_end.tms_stime) / CLK_TCK,
	  (float)((t_end.tms_utime - t_eppstein_end.tms_utime)
		  + (t_end.tms_stime - t_eppstein_end.tms_stime)) / CLK_TCK);
  fprintf(stderr, "  Total            %6.2f  %6.2f  %6.2f\n",
	  (float)(t_end.tms_utime - t_start.tms_utime) / CLK_TCK,
	  (float)(t_end.tms_stime - t_start.tms_stime) / CLK_TCK,
	  (float)((t_end.tms_utime - t_start.tms_utime)
		  + (t_end.tms_stime - t_start.tms_stime)) / CLK_TCK);
#elif defined(sun)
  fprintf(stderr, "  Pre-process      %6.2f  %6.2f  %6.2f\n",
	  (float)(t_search_start.tms_utime - t_start.tms_utime) / HZ,
	  (float)(t_search_start.tms_stime - t_start.tms_stime) / HZ,
	  (float)((t_search_start.tms_utime - t_start.tms_utime)
		  + (t_search_start.tms_stime - t_start.tms_stime)) / HZ);
  fprintf(stderr, "  Search(opt)      %6.2f  %6.2f  %6.2f\n",
	  (float)(t_optimal_end.tms_utime - t_search_start.tms_utime) / HZ,
	  (float)(t_optimal_end.tms_stime - t_search_start.tms_stime) / HZ,
	  (float)((t_optimal_end.tms_utime - t_search_start.tms_utime)
		  + (t_optimal_end.tms_stime - t_search_start.tms_stime))
	  / HZ);
  fprintf(stderr, "  Search(total)    %6.2f  %6.2f  %6.2f\n",
	  (float)(t_search_end.tms_utime - t_search_start.tms_utime) / HZ,
	  (float)(t_search_end.tms_stime - t_search_start.tms_stime) / HZ,
	  (float)((t_search_end.tms_utime - t_search_start.tms_utime)
		  + (t_search_end.tms_stime - t_search_start.tms_stime))
	  / HZ);
  fprintf(stderr, "  Eppstein Process %6.2f  %6.2f  %6.2f\n",
	  (float)(t_eppstein_end.tms_utime - t_search_end.tms_utime) / HZ,
	  (float)(t_eppstein_end.tms_stime - t_search_end.tms_stime) / HZ,
	  (float)((t_eppstein_end.tms_utime - t_search_end.tms_utime)
		  + (t_eppstein_end.tms_stime - t_search_end.tms_stime))
	  / HZ);
  fprintf(stderr, "  Post-process     %6.2f  %6.2f  %6.2f\n",
	  (float)(t_end.tms_utime - t_eppstein_end.tms_utime) / HZ,
	  (float)(t_end.tms_stime - t_eppstein_end.tms_stime) / HZ,
	  (float)((t_end.tms_utime - t_eppstein_end.tms_utime)
		  + (t_end.tms_stime - t_eppstein_end.tms_stime)) / HZ);
  fprintf(stderr, "  Total            %6.2f  %6.2f  %6.2f\n",
	  (float)(t_end.tms_utime - t_start.tms_utime) / HZ,
	  (float)(t_end.tms_stime - t_start.tms_stime) / HZ,
	  (float)((t_end.tms_utime - t_start.tms_utime)
		  + (t_end.tms_stime - t_start.tms_stime)) / HZ);
#else
  fprintf(stderr, "  Pre-process      %6.2f  %6.2f  %6.2f\n",
	  (float)(t_search_start.tms_utime - t_start.tms_utime),
	  (float)(t_search_start.tms_stime - t_start.tms_stime),
	  (float)((t_search_start.tms_utime - t_start.tms_utime)
		  + (t_search_start.tms_stime - t_start.tms_stime)));
  fprintf(stderr, "  Search(opt)      %6.2f  %6.2f  %6.2f\n",
	  (float)(t_optimal_end.tms_utime - t_search_start.tms_utime),
	  (float)(t_optimal_end.tms_stime - t_search_start.tms_stime),
	  (float)((t_optimal_end.tms_utime - t_search_start.tms_utime)
		  + (t_optimal_end.tms_stime - t_search_start.tms_stime))
	 );
  fprintf(stderr, "  Search(total)    %6.2f  %6.2f  %6.2f\n",
	  (float)(t_search_end.tms_utime - t_search_start.tms_utime),
	  (float)(t_search_end.tms_stime - t_search_start.tms_stime),
	  (float)((t_search_end.tms_utime - t_search_start.tms_utime)
		  + (t_search_end.tms_stime - t_search_start.tms_stime))
	 );
  fprintf(stderr, "  Eppstein Process %6.2f  %6.2f  %6.2f\n",
	  (float)(t_eppstein_end.tms_utime - t_search_end.tms_utime),
	  (float)(t_eppstein_end.tms_stime - t_search_end.tms_stime),
	  (float)((t_eppstein_end.tms_utime - t_search_end.tms_utime)
		  + (t_eppstein_end.tms_stime - t_search_end.tms_stime))
	 );
  fprintf(stderr, "  Post-process     %6.2f  %6.2f  %6.2f\n",
	  (float)(t_end.tms_utime - t_eppstein_end.tms_utime),
	  (float)(t_end.tms_stime - t_eppstein_end.tms_stime),
	  (float)((t_end.tms_utime - t_eppstein_end.tms_utime)
		  + (t_end.tms_stime - t_eppstein_end.tms_stime)));
  fprintf(stderr, "  Total            %6.2f  %6.2f  %6.2f\n",
	  (float)(t_end.tms_utime - t_start.tms_utime),
	  (float)(t_end.tms_stime - t_start.tms_stime),
	  (float)((t_end.tms_utime - t_start.tms_utime)
		  + (t_end.tms_stime - t_start.tms_stime)));
#endif
#endif

  if ((opt_sols != NULL)||(opt_sols2 != NULL)||(opt_sols3 != NULL)) {
    node_set_finish(node_set);
  }
  div_score();
  return;
}
