#include <stdio.h>
#include <stdlib.h>
#define BUFSIZE 128
#define DEFAULT_DIM 16
#define MAX_DIM 128

int **weights;
int dim;

int takeword(FILE *fp,char *buf);
void read_weight_file(FILE *w_fp);
void print_weights();
int gcd_primitive(int a, int b);
int gcd(int a, int b);
int normalize_weights();
void help();

int takeword(FILE *fp,char *buf)
{
  char c;
  int i;

  c = fgetc(fp);
  while ((c == ' ')||(c == '\t')||(c == '\n')||(c == '=')
	 ||(c == ',')) c = fgetc(fp);
  i = 0;
  while (1) {
    if (c == EOF) {buf[i] = '\0';return -1;}
    if ((c == ' ')||(c == '\t'))
      {buf[i] = '\0';return 0;}
    if (c == '\n') {buf[i] = '\0';return 1;}
    buf[i] = c;
    c = fgetc(fp);
    i++;
  }
}

void read_weight_file(FILE *w_fp)
{
  int i, j, newline;
  char buf[BUFSIZE];

  for(i=0;i<dim;i++) {
    for(j=0;j<dim;j++) {
      newline = takeword(w_fp, buf);
      sscanf(buf,"%d",&(weights[i][j]));
    }
    while (!newline) {newline = takeword(w_fp, buf);}
  }
}

void print_weights(FILE *wfp)
{
  int i,j;

  for(i=0;i<dim;i++) {
    for(j=0;j<dim;j++) {
      fprintf(wfp, " %3d", weights[i][j]);
    }
    fprintf(wfp, "\n");
  }
  printf("\n");
}

int gcd_primitive(int a, int b)   /* must be a >= b >= 0 */
{
  if (b == 0) return (a);
  return gcd_primitive(b, (a % b));
}

int gcd(int a, int b)
{
  if (a < 0) a = -a;
  if (b < 0) b = -b;
  if (a < b) return gcd_primitive(b, a);
  else return gcd_primitive(a, b);
}

int normalize_weights()
{
  int i,j;
  int d;

  d = weights[0][0];
  for(i=0;(i<dim-1);i++) {
    for(j=i+1;j<dim;j++) {
      d = gcd(weights[i][j],d);
      if (d == 1) return 1;
    }
  }
  for(i=0;i<dim;i++)
    for(j=0;j<dim;j++) {
      weights[i][j] /= d;
    }
  return d;
}

void help()
{
  printf(" 0: make 0\n");
  printf(" a: Add importance of one sequence\n");
  printf(" c: Change dimension\n");
  printf(" d: Divide importance of one sequence\n");
  printf(" h: Show this message\n");
  printf(" i: Initialize\n");
  printf(" m: Multiply importance of one sequence\n");
  printf(" n: Normalize weight table\n");
  printf(" o: Output to file\n");
  printf(" p: Put x to weights[i][j]\n");
  printf(" q: Quit this program\n");
  printf(" r: Read from file\n");
  printf(" s: Show current weight table\n");
}

void main(int argc, char *argv[])
{
  int i, j;
  int s, m, nd;
  char buf[BUFSIZE], buf2[BUFSIZE];
  FILE *w_fp;

  if(argc < 2) {
    dim = DEFAULT_DIM;
    printf("Used default dimension: %d\n", dim);
  } else {
    dim = atoi(argv[1]);
    if (dim <= 0) {
      printf("Dimension too large\n");
      dim = DEFAULT_DIM;
      printf("Used default dimension: %d\n", dim);
    }
    if (dim >= MAX_DIM) {
      printf("Dimension too small\n");
      dim = DEFAULT_DIM;
      printf("Used default dimension: %d\n", dim);
    }
  }
  if ((weights = (int **)malloc(dim*sizeof(int *))) == NULL) {
    fprintf(stderr,"Could not allocate memory for wight matrix\n");
    exit(1);
  }
  for (i = 0;i < dim;i++) {
    if ((weights[i] = (int *)malloc(dim*sizeof(int))) == NULL) {
      fprintf(stderr,"Could not allocate memory for wight matrix\n");
      exit(1);
    }
    for (j = 0;j < dim;j++) {
      if (i==j) weights[i][j] = 0;
      else weights[i][j] = 1;
    }
  }

  printf("Initial Weight Table:\n");
  print_weights(stdout);

  while(1) {
    printf("\n> ");
    scanf("%s",buf);
    switch(buf[0]) {
    case '0':
      for(i=0;i<dim;i++) for(j=0;j<dim;j++) {
	weights[i][j] = 0;
      }
      printf("New Weight Table:\n");
      print_weights(stdout);
      break;
    case 'a':
      printf(" Sequence: ");
      scanf("%d",&s);
      printf(" Adder: ");
      scanf("%d",&m);
      if (s<0) s = 0;
      if (s>=dim) s = dim-1;
      for(i = 0; i < dim;i++) {
	if (s != i) {
	  weights[s][i] += m;
	  weights[i][s] += m;
	}
      }
      printf("New Weight Table:\n");
      print_weights(stdout);
      break;
    case 'c':
      printf(" Change dimension (<%d) > ", dim);
      scanf("%d",&nd);
      if (nd <= 1) nd = DEFAULT_DIM;
      if (nd < dim) {
	dim = nd;
	printf("New Weight Table:\n");
	print_weights(stdout);
      }
      else printf(" Not updated dimension\n");
      break;
    case 'd':
      printf(" Sequence: ");
      scanf("%d",&s);
      printf(" Divider: ");
      scanf("%d",&m);
      if (s<0) s = 0;
      if (s>=dim) s = dim-1;
      for(i = 0; i < dim;i++) {
	weights[s][i] /= m;
	weights[i][s] /= m;
      }
      printf("New Weight Table:\n");
      print_weights(stdout);
      break;
    case 'h':
      help();
      break;
    case 'i':
      for(i=0;i<dim;i++) for(j=0;j<dim;j++) {
	if (i==j) weights[i][j] = 0;
	else weights[i][j] = 1;
      }
      printf("New Weight Table:\n");
      print_weights(stdout);
      break;
    case 'm':
      printf(" Sequence: ");
      scanf("%d",&s);
      printf(" Multiplier: ");
      scanf("%d",&m);
      if (s<0) s = 0;
      if (s>=dim) s = dim-1;
      for(i = 0; i < dim;i++) {
	weights[s][i] *= m;
	weights[i][s] *= m;
      }
      printf("New Weight Table:\n");
      print_weights(stdout);
      break;
    case 'n':
      normalize_weights();
      printf("New Weight Table:\n");
      print_weights(stdout);
      break;
    case 'o':
      printf(" Output file: ");
      scanf("%s",buf2);
      if((w_fp = fopen(buf2, "w")) == NULL){
	fprintf(stderr, "Can't open %s\n", buf2);
	break;
      }
      print_weights(w_fp);
      fclose(w_fp);
      break;
    case 'p':
      printf(" Put x to weights[i][j] (x/i/j) :");
      scanf("%d %d %d", &m, &i, &j);
      weights[i][j] = m;
      weights[j][i] = m;
      printf("New Weight Table:\n");
      print_weights(stdout);
      break;
    case 'q':
/*      printf(" Really want to quit? (y/n) ");
      scanf("%s", buf2);
      if (buf2[0] == 'y') {*/
	printf("Have a nice day!\n");
	exit(0);
/*      }*/
    case 'r':
      printf(" Weight File Name: ");
      scanf("%s",buf2);
      if((w_fp = fopen(buf2, "rt")) == NULL){
	fprintf(stderr, "Can't open %s\n", buf2);
	break;
      }
      read_weight_file(w_fp);
      fclose(w_fp);
      printf("New Weight Table:\n");
      print_weights(stdout);
      break;
    case 's':
      printf("Current Weight Table:\n");
      print_weights(stdout);
      break;
    }
  }
}
